--[[
SUGARSCAPE MODEL 
Based on the book J. Epstein and R. Axtell, "Growing artificial societies: social science from the bottom up".

Implemented in TerraME by Gilberto Câmara (December, 2013)
Based on earlier version by Pedro Andrade (June, 2013)

Sugarscape is an abstract agent-based model of artificial societies developed by 
Epstein and Axtell to investigate the emergence of social processes.

The model represents an artificial society in which agents move over a 50 × 50 cell grid. 
Each cell has a gradually renewable quantity of ‘sugar’, which the agent located at that cell can eat. 
However, the amount of sugar at each location varies. Agents have to consume sugar in order to survive.
If they harvest more sugar than they need immediately, they can save it and eat it later 
(or, in more complex variants of the model, can trade it with other agents). 

Agents can look to the north, south, east and west of their current locations and can see a distance 
which varies randomly, so that some agents can see many cells away while others can only see adjacent cells.

Agents also differ in their ‘metabolic rate’, the rate at which they use sugar. 
If their sugar level ever drops to zero, they die. 
In some simulations, new agents replace the dead ones with a random initial allocation of sugar. 

Thus there is an element of the ‘survival of the fittest’ in the model, since those agents 
that are relatively unsuited to the environment because they have high metabolic rates, 
poor vision, or are in places where there is little sugar for harvesting, die relatively 
quickly of starvation. 

Epstein and Axtell (1996) present a series of elaborations of this basic model 
in order to illustrate a variety of features of societies. The basic model shows that even 
if agents start with an approximately symmetrical distribution of wealth (the amount of sugar 
each agent has stored), a strongly skewed wealth distribution soon develops. This is because 
a few relatively well-endowed agents are able to accumulate more and more sugar, while the majority 
only barely survive or die.
]]

--[[ 
 	 This replication assumes you have Epstein and Axtell's book
	 Please also read and download Anthony Bigbee MsC thesis 
	 "Replication of Sugarscape using MASON"
	 http://mason-sugarscape.googlecode.com/svn-history/r22/trunk/docs/ajb_thesis_revised.pdf
]]
-- Auxiliary modules
tmepath = os.getenv("TME_PATH_1_3_0")
package.path = tmepath.."/contrib/terrameaddons/?.lua;"..tmepath.."/contrib/viewerlib/?.lua;"..package.path
local rand  = require ("random")
local cells = require ("cellspace_functions" )
local agts  = require ("agent_functions")
local view  = require ("viewer")
local csv   = require ("csv")
local stats = require ("stats")
local util  = require ("util")
local gris  = require ("grissom")

local rules = require ("rules")

-- Defines the default "Sugarscape" model

-- Basic parameters 
DIM_SPACE = 50    -- dimension of the cell space
NUM_RUNS  = 500   -- number of iterations

-- Default model parameters
__Sugarscape = {
	-- name of the file with the initial values of the sugarscape
	sugarscapeFile           = "sugar_map.txt",
	-- model keeps track of the current time 
	currentTime              =  0,
	-- dimension of cell space 
	dimSpace                 =  DIM_SPACE,
	-- number of iterations
	numRuns                  =  NUM_RUNS, 
	-- number of agents          
	numAgents                =  600,           
	-- initial sugar available per agent
	agentWealth              =  { min = 5,  max = 25 }, 
	-- consumption per agent per time interval
	agentMetabolism          =  { min = 1, max = 4   },
	-- vision in horizontal and vertical directions
	agentVision              =  { min = 1, max = 6   },
	-- agent lifetime (by default, agents have no lifetime limits)
	agentLifetime            =  { min = math.huge, max = math.huge},
      -- minimum sugar per cell
	cellSugar                =  { min = 0,  max = 4},
	-- reference to cell's von Neumann neighborhood
	csVonNeumanNeighborhood  = "neigh1",
	-- normal growth rate for sugar
	growthRate               =  1, 
	-- pollution coeficients
	pollutionProductionRate  =  0,
	pollutionConsumptionRate =  0,
	pollutionStartTime       =  math.huge,
	diffusionStartTime       =  math.huge,
	-- seasonal duration?
	hasSeasons               =  false,
	seasonDuration           =  math.huge,
	-- block (if agents are placed in blocks)
	block                    =  { xmin = 0, ymin = 0, xmax = DIM_SPACE - 1, ymax = DIM_SPACE - 1 },

	-- Agent and Sugarscape Rules 
	-- how agents are placed in the cell space
	placementRule            =  rules.randomPlacement,
	-- how agents move in the cell space
	movementRule             =  rules.gradientSearch,
	-- search maximization Criteria
	searchMaxRule            =  rules.maxSugar,
	-- how agents' metabolism works
	metabolismRule           =  rules.eatAllSugar,
	-- how agents are replaced if the die
	replacementRule          =  rules.noReplacement,
	-- how the agents pollute the space
	pollutionFormationRule   =  rules.noPollution,
	pollutionDiffusionRule   =  rules.noPollution,
	-- how sugar grows back in the cell space
	growbackRule             =  rules.normalGrowth,
	-- how social networks are handled
	socialNetworkRule        =  rules.noSocialNetworks,
	-- model visualization
	viewRule                 =  rules.updateViews,

	-- Visualization parameters and functions
	-- show how numagents evolve?
	showNumAgents            =  false,
	-- show Gini coefficient?    
	showGiniIndex            =  false,
	-- show social networks?
	showSocialNetworks       =  false,
	-- show wealth distribution?	   
	showWealthDist           =  false, 
	-- how to show the agents in the sugarscape?
	-- agentColor               =  5, 
	-- how to show the social network in the sugarscape?
	socialNetworkColor       =  3,
	-- show original sugarscape?
	showOriginalSugarscape   =  true,
	-- file to save histogram      
	distFile                 =  "sugarscape_dist",  
	-- used to wait some time before next iteration
	viewWait                 =  10,
    
    -- functions
    -- create the cell space for sugarscape       
    makeCS  =   function (self) 
                    -- read Sugarscape file - approximates the original Epstein and Axtell grid 
                    self.cs = cells.fromTxt {
                        xdim        = model.dimSpace, 
                        ydim        = model.dimSpace, 
                        attrname    = "maxsugar", 
                        txtfile     = "sugar-map.txt", 
                        charsize    = 1, 
                        whitespaces = 1
                        }
                    assert (self.cs ~= nil )
                    -- create different rook neighborhoods for agents of different visions
                    for i = self.agentVision.min, self.agentVision.max do
                        cells.createRookNeighborhood(self.cs, i, self.dimSpace)
                    end
                    -- initialize the sugar and color attributes of the cell 
                    forEachCell(self.cs, 
                        function(cell)
                            cell.sugar         = cell.maxsugar
                            cell.color         = cell.maxsugar
                            cell.socialNetwork = 0
                            cell.harvestTime   = 0
                            cell.production    = 0
                            cell.consumption   = 0 
                            cell.pollution     = 0
                        end
                    )

                    if (self.hasSeasons) then
                        self.north, self.south = cells.northSouthTrajectories (self.cs)
                    end
                    self.cs:synchronize("pollution")
                end,
                
    
    --[[ create a agent 
         In Sugarscape, the agent has attributes: age, wealth (the amount of sugar), 
            life-expectancy (the maximum age that can be reached), 
            metabolism (how much sugar an agent eats each time period), and  
            vision (how many cells ahead an agent can see). 
            The agent’s life-expectancy, metabolism and vision do not change, 
            while age and wealth do change.
      --]]
    makeAgent = function (self)
                    self.agent = Agent {
                    init = function(ag)
                        ag.wealth     = rand:integer (self.agentWealth.min,     self.agentWealth.max   )
                        ag.metabolism = rand:integer (self.agentMetabolism.min, self.agentMetabolism.max)
                        ag.maxage     = rand:integer (self.agentLifetime.min,   self.agentLifetime.max  )
                        ag.vision     = rand:integer (self.agentVision.min,     self.agentVision.max    )
                        ag.age        = 0 
                        ag.sn         = SocialNetwork()
                        ag.placed     = 1
                        end,
                    execute = function(ag) 
                        end
                    }
                    assert (self.agent ~= nil )
                end,

    -- Creates a society of agents 
    makeSoc =   function (self)
                    self.society = Society {
                                        instance    = self.agent, 
                                        quantity    = self.numAgents,
                                        wealthHist  = {},
                                        histValues  = {},
                                        giniIndex   = 0
                                        }
                    assert (self.society ~= nil )
                   
                end,
                    
    -- Place the society of agents in the cell space
    makeEnv =   function (self) 
                    assert (self.society ~= nil )
                    assert (self.cs ~= nil )
                    self.env = Environment { self.cs, self.society}

                    self.env:createPlacement { strategy = "void" }
                    self:placementRule()
                end,
                        
    -- create model visualisations                    
    makeViews = function (self)
                    self:viewCellSpace()        
                    self:viewSociety()
                    self.plot = Cell {
                        giniIndex = 0,
                        nagents   = self.numAgents
                    }
                    if (self.showGiniIndex) then
                        self:giniPlot ()
                    end
                    if (self.showNumAgents) then
                        self:nagtsPlot ()
                    end
                    if (self.showSocialNetworks) then
                        self:viewSN ()
                    end

                end,
                    
    -- create a gini plot                 
    giniPlot =  function (self)
                    self.giniPlotObs = view.Plot {
                                        subject     = self.plot,
                                        attributes  = {"giniIndex"}
                                    }
                    end,
                    
    -- create a plot with the number of agents 
    nagtsPlot = function (self)
                    self.nagentsObs = view.Plot {
                                        subject     = self.plot,
                                        attributes  = {"nagents"},
                                        title       = "Agents in the model"
                                    } 
                    end,

    -- create a map to view the cell space
    viewCellSpace =     function (self)
                            assert (self.cs ~= nil )
                            self.mapcs = view.Map {
                                    subject    = self.cs,
                                    attributes = {"sugar"},
                                    grouping   = "uniquevalue",
                                    values     = {0, 1, 2, 3, 4, 5, 6},
                                    slices     = 8,
                                    minimum    = self.cellSugar.min,
                                    maximum    = self.cellSugar.max,
                                    colors     = "Blues"
                                    }
                        end,
                
    -- create a map to view the agents      
     viewSociety =  function (self)
                        forEachAgent (self.society, 
                                        function (ag) ag.obs = self.viewAg (ag, self.mapcs) 
                                        end 
                                    )
                    end,
     viewAg =   function (ag, obsCS)
                    local obs = view.Symbol {
                                subject     = ag,
                                observer    = obsCS,
                                attributes  = {"metabolism"},
                                grouping    = "uniquevalue",
                                values      = {0, 1, 2, 3, 4},
                                font        = "Grissom Free",
                                fontSize    = 18,
                                colors      = "YlOrBr",
                                symbol      = gris.FLY
                    }
                    return obs
                end,
                            
    --  Create a map to view the social networks
    viewSN =    function (self)
                    assert (self.cs ~= nil )
                    self.mapsn = view.Map {
                                    subject   = self.cs,
                                    attributes = {"socialNetwork"},
                                    values    = {0, 1, 2, 3},
                                    colors    = "Reds"
                                    }
                end,
    
   
        

    -- save the wealth distribution histogram 
    wealthDist = function (self)
                    if (self.showWealthDist) then
                        dist = stats.distrib (agts.attrValues(model.society, "wealth"))
                        fn   = "./histog/"..self.distFile .. "_" .. self.currentTime ..".csv"
                        csv.saveDistrib (fn, dist)
                    end 
                end,

    -- creates the timer and defines how the model is executed
    makeTimer = function (self)
                    self.timer = Timer {
                                Event {time = 1, period = 1, priority = 1, 
                                    action = function(ev)
                                                self.currentTime = ev:getTime()
                                                self:movementRule()
                                                self:metabolismRule()
                                                self:pollutionFormationRule()
                                                self:pollutionDiffusionRule()
                                                self:replacementRule()
                                                self:socialNetworkRule()
                                                self:growbackRule()
                                                self:viewRule()
                                                util.wait (self.viewWait)
                                            end
                                },
                                Event {time = 1, period = 50, priority = 2, 
                                            action = function(ev)
                                                self:wealthDist()
                                            end
                                }
                    }
                end,

    -- Step by step creation of the model
    setup   =   function (self)
                    -- builds the cell space 
                    self:makeCS () 
                    -- defines agent behaviour     
                    self:makeAgent ()
                    -- creates a society        
                    self:makeSoc () 
                    -- place the society in the cell space       
                    self:makeEnv ()  
                    -- create visualisations
                    self:makeViews () 	 
                    -- control model execution        
                    self:makeTimer ()          
                end,
            
    -- runs the sugarscape model

    run     =   function (self)
                    -- set up the model
                    self:setup ()
                    -- execute the model 
                    self.timer:execute (NUM_RUNS)
                end
}



--[[ To build the sugarscape scenarios, we use Lua's metatable.
     Metatables allow a table to use values from another one. 
     In our case, we define a metatable (Sugarscape) that is common
     to all sugarscape scenarios. The metatable Sugarscape contains
     all the default values needed by the different sugarscape models.

     Each sugarscape scenario only defines what the parameters and rules 
     that are different from defaults ]]

-- defines that Sugarscape (see above) is a metatable 
metaTableSugarscape = {__index = __Sugarscape}

--[[ This function is used to implement all sugarscape models ]]
function Sugarscape(argv)
	-- the model first gets the scenario-specific parameters
	local model = argv
	-- all other parameters are accessed via the metatable
	setmetatable(model, metaTableSugarscape)
return model
end



