
--[[ Sugarscape scenarios - parameters and rules that are different from defaults ]]

tmepath = os.getenv("TME_PATH_1_3_0")
package.path = tmepath.."/contrib/terrameaddons/?.lua;"..tmepath.."/contrib/viewerlib/?.lua;"..package.path

local rules  = require ("rules")
local _      = require ("moses")
require ("sugarscape")
-- require ("mobdebug").start("localhost", 8172)

--[[ Animation II-1 - immediate growth, default movement and metabolism
	For this first run, we take the initial population to be 400
	agents arranged in a random spatial distribution. 
	For the first run, the sugarscape will follow "immediate growback rule".
	The sugarscape grows back to full capacity immediately.
]]
sc_II_1 = Sugarscape {
	growbackRule    =  rules.immediateGrowth,
	showNumAgents   =  true
}

--[[ Animation II-2 - normal growth, default movement and metabolism
	For this second run we again take the initial population to be 400
	agents arranged in a random spatial distribution. 
	Each agent again executes the default rules for movement and metabolism. 
	But now let us change the sugarscape growth rule: Every site
	whose level is less than its capacity grows back at 1 unit per time period. 
]]
sc_II_2 = Sugarscape {
	growbackRule    =  rules.normalGrowth,
	showNumAgents   =  true,
	showGiniIndex   =  true,
	showWealthDist  =  false,
}
--[[ Animation II-3 - normal growth, agent replacement, default movement and metabolism
	 
	 Shows how giniIndex and wealth distribution evolve 
	 
	 We set each agent's maximum achievable age- beyond which it cannot live - 
	 to a random number drawn from some interval [a,b]. Of course, agents can still die 
	 of starvation, as before.
	 Given that agents are to have finite lifetimes, the second modification
	 that must be implemented is a rule of agent replacement. 
	 However, to ensure a stationary wealth distribution it is desirable to use a 
	 replacement rule that produces a constant population .
	
	 When an agent dies it is replaced by an agent of age 0 having random genetic attributes, 
	 random wealth and position on the sugarscape, and a maximum age randomly selectedfrom the range [a,b].

	 We want to track the distribution of wealth, and we show a histogram of wealth. 
	 While initially quite symmetrical, the distribution ends up highly skewed.
	 Such skewed wealth distributions are produced for wide ranges of agent and environment 
	 specifications. They seem to be characteristic
	 of heterogeneous agents extracting resources from a landscape of fixed capacity.

	 The animation also displays a real-time computation of the Gini coefficient. 
	 Note that it starts out quite small (about 0.230) and ends up fairly large (0.500). 
]]
sc_II_3 = Sugarscape {
	numAgents            =  250,
	growbackRule         =  rules.normalGrowth,
	replacementRule      =  rules.ageReplacement,         
	agentLifetime        =  { min = 60, max = 100},
	showNumAgents        =  true,
	showGiniIndex        =  true,
	showWealthDist       =  true
}
--[[ Animation II-4 is the same as II-3 ]]

--[[ Animation II-5 - normal growth, default movement and metabolism, showing socialnetworks
	We want to keep track of each agent's neighbors.
	In what follows we shall always employ the von Neumann neighborhood.
	When an agent moves to a new position on the sugarscape it has
	from zero to four neighbors. Each agent keeps track of these neighboring
	agents internally until it moves again, when it replaces its old neighbors 
	with its new neighbors.
]]
sc_II_5 = Sugarscape {
	growbackRule         =  rules.normalGrowth,
	showNumAgents        =  true,
	showSocialNetworks   =  true,
	socialNetworkRule    =  rules.buildSocialNetworks
}
--[[
	Animation II-6 - initial distribution of agents in a block

	We turn now to a different kind of emergent structure, this one
	spatial in nature.  Instead of the random initial distribution of agents 
	on the sugarscape used earlier, suppose they are initially clustered in the dense block.
      In all other respects the agents and sugarscape are exactly as in animation II-2. 
      How will this block start affect the dynamics? A succession of coherent waves results, 
      a phenomenon we did not expect.

	N.B: This animation is not reproducible directly following the descriptions of the E&A book.
	GROWBACK must be delayed by one time step (sugar does not growback as it is consumed)
	MIN_VISION must be increased to 5 
	MAX_VISION must be increased to 15
]]
sc_II_6 = Sugarscape {
	block               =  {xmin = 0, ymin = DIM_SPACE - 20, xmax = 19, ymax = DIM_SPACE - 1 },
	placementRule       =  rules.blockPlacement,
	growbackRule        =  rules.delayedGrowth,
	agentVision         =  {min = 5, max = 15}
}

--[[ Animation II-7 Seasonal growback rule
     Initially it is summer in the top half of the sugarscape and winter in the bottom half.
     Then, every Y time periods the seasons flip - in the region where it was summer it becomes
     winter and vice versa. For each site, if the season is summer then sugar grows back at a 
     rate of A units per time interval; if the season is winter then the growback rate 
     is A units per B time intervals
]]
sc_II_7 = Sugarscape {
	showNumAgents       =  true,
	hasSeasons      	=  true,
	growbackRule        =  rules.seasonalGrowth,
	seasonDuration   	=  50,
	summerGrowthRate    =  1,
	winterGrowthRate    =  0.125
}

--[[ 
	Animation II-8: Pollution in the sugarscape 

	Pollution formation rule: when sugar S is gathered from the sugarscape, an amount of 
	production pollution is generated in quantity ALPHA*S. When sugar amount M is metabolized, 
	consumption pollution is generated according to BETA*M. 
	The total pollution on a site at time t, P(T), is the sum of the
	pollution present at the previous time, plus the pollution
	resulting from production and consumption activities.

	Pollution diffusion rule: Diffusion on a sugarscape is simply
	implemented as a local averaging procedure. 
	That is, diffusion transports pollution from sites of high levels to sites of low levels.

	The new agent movement rule modified for pollution is to
	look out as far as vision permits in the four principal lattice
	directions and identify the unoccupied site(s) having the maximum sugar to pollution ratio.
]]

sc_II_8 = Sugarscape {
	showNumAgents              = true,
	searchMaxRule              = rules.maxSugarToPollution,  	-- search maximization Criteria
	pollutionProductionRate    = 1,
	pollutionConsumptionRate   = 1,
	pollutionStartTime         = 50,
	diffusionStartTime         = 100,
	-- how the agents pollute the space
	pollutionFormationRule     = rules.pollutionProdCons,
	pollutionDiffusionRule     = rules.pollutionLocalDiffusion
} 

--[[ 
	We now can choose our model from the available scenarios
]]
MODEL_SCENARIOS   = {sc_II_1, sc_II_2, sc_II_3, sc_II_5, sc_II_6, sc_II_7, sc_II_8 }

-- Choose from one of the above
model             =  sc_II_2
assert (_.contains (MODEL_SCENARIOS, model))

-- Run the model!!
model:run ()
